import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Random;

/**
 * @author Esteve Mir Esprrech
 * v1.0 November 2015
 * This class represents an individual ant agent for MMAS algorithm 
 */
public class ant {
	
	private int antID;
	private double alpha; 				//Pheromone trail importance
	private double beta;  				//Heuristic function importance
	private double exTh;				//Exploration threshold. Exploration vs explotation
	private double exThRandom;			//Random between 0 and 1
	private List<flight> flights;		//Flights in problem
	private int [] nodesFlights;
	private List<flight> antFlights;	//Original flights
	private double [][] trail;			//Trails of pheromones between edges
	private double [][] heuristic;		//Heuristic function between edges
	private List<List<flight>> pairings;	//Generated pairings of flights
	private Boolean finished = false;
	
	/**
	 * Constructor
	 * 
	 * @param ID:
	 * 		Agent ID
	 * @param alpha:
	 * 		Pheromone modifier
	 * @param beta:
	 * 		Heuristic modifier
	 * @param exTh:
	 * 		Transition State function control parameter
	 * @param flights:
	 * 		Number of flights
	 * @param trail:
	 * 		Pheromone trails for arcs in the problem
	 * @param heuristic:
	 * 		Heuristic values for arcs in the problem
	 */
	public ant (int ID, double alpha, double beta, double exTh, List<flight> flights, double[][] trail, double[][] heuristic){
		Random generator = new Random();
		
		this.antID = ID;
		this.alpha = alpha;
		this.beta = beta;
		this.exTh = exTh;
		this.exThRandom = generator.nextDouble(); //Initialize every ant with a random exploration/explotation [0-1]
		this.flights = flights;
		this.antFlights = new ArrayList<flight>(this.flights);
		Collections.copy(antFlights, flights);
		this.nodesFlights = new int[flights.size()];
		this.trail = trail;
		this.heuristic = heuristic;
		pairings = new ArrayList<List<flight>>();
	}
	
/*************************************MAIN FUNCTION***********************************************************/	
	/**
	 * Init function for the agent
	 */
	public void init(){
		
		int presentPairing = 0;
		this.pairings.add(new ArrayList<flight>());
		flight presentFlight = antFlights.get(getEarliestFlight(this.flights));  //Begin on the earliest flight
		this.pairings.get(presentPairing).add(new flight(presentFlight)); 
		this.nodesFlights[presentFlight.getID()] = 1;
		List <flight> tempFlights = new ArrayList<flight>();
		List <flight> pendingFlights = new ArrayList<flight>();
		
		
		//While all the nodes have not been visited yet do
		while (!checkVisitedNodes()){
			
			//Get all the flights that have been not visited yet
			pendingFlights.clear();
			pendingFlights = getPendingFlights();
			
			//Get all the flights that comply with the constrains compared to the actual selected flight
			tempFlights.clear();
			tempFlights = getConstrainedFlights(presentFlight, pendingFlights, this.pairings.get(presentPairing)); //Get flights that comply with constraints
			
			//Use q and q0 values (exThRandom and exTh) to select between explotation and exploration
			if (!tempFlights.isEmpty())
			{
				int lastFlightID = presentFlight.getID();
				
				if (this.exThRandom<this.exTh){
					presentFlight = tempFlights.get(transitionState(lastFlightID, tempFlights));
					this.pairings.get(presentPairing).add(new flight(presentFlight));
					this.nodesFlights[presentFlight.getID()] = 1;
				} else {
					presentFlight = tempFlights.get(computeTotalProbability(lastFlightID, tempFlights));
					this.pairings.get(presentPairing).add(new flight(presentFlight));
					this.nodesFlights[presentFlight.getID()] = 1;
				}
				
			} 
			//If there are no flights that comply with the constraints, generate a new pairing
			else if (!checkVisitedNodes())  
			{
				this.pairings.add(new ArrayList<flight>());
				presentFlight = pendingFlights.get(getEarliestFlight(pendingFlights));
				presentPairing++;
				this.pairings.get(presentPairing).add(new flight(presentFlight));
				this.nodesFlights[presentFlight.getID()] = 1;
			} 	 
		}
		this.finished = true;
	}

/**********************************************FUNCTIONALITIES**************************************************/	

	
	/**
	 * Choose a flight based on the best probability between the present flight
	 * and the next one.
	 * @param presentFlightID: The ID of the present flight
	 * @param remainingFlights:
	 *			List containing all the remaining flights
	 * @return:
	 * 			The position of the next best to choose flight in the remaining list of flights to choose
	 */
	private int transitionState(int presentFlightID, List<flight> remainingFlights){
		double bestProb=0,actualProb=0;
		int index = 0, finalIndex=0;
		int nextID;
		for (flight x: remainingFlights){
			nextID = x.getID();
			actualProb = Math.pow(this.trail[presentFlightID][nextID], this.alpha)*Math.pow(this.heuristic[presentFlightID][nextID], this.beta);

			if (actualProb>bestProb){
				bestProb = actualProb;
				finalIndex = index;
			}
			index++;
		}
		
		return finalIndex;
	}
	
	/**
	 * * Choose a flight based on its computated probability.
	 * It uses a probability equation which depends on the:
	 * - heuristic value for edges
	 * - alpha and beta values 
	 * @param presentFlightID: 
	 * 			Represents the node on which the ant is
	 * @param remainingFlights:
	 * 			List containing all the remaining flights
	 * @return:
	 * 			The position of the next best to choose flight in the remaining list of flights to choose
	 */
	private int computeTotalProbability(int presentFlightID, List<flight> remainingFlights){
		Random gen = new Random();
		int nextID;
		int index=0;
		double [] probabilities = new double[remainingFlights.size()];
		double totalProbability =0;
		double randChoose = gen.nextInt(101);
		
		
		for (flight x: remainingFlights){
			nextID = x.getID();
			probabilities[index] = Math.pow(this.trail[presentFlightID][nextID], this.alpha)*Math.pow(this.heuristic[presentFlightID][nextID], this.beta);
			totalProbability += probabilities[index];
			index++;
		}
		for (int i=0; i<probabilities.length; i++){
			if(i==0) 	probabilities[i] = (probabilities[i]*100/totalProbability);
			else 		probabilities[i] = (probabilities[i]*100/totalProbability)+probabilities[i-1];
		}
		for (int i=0; i<probabilities.length; i++){
			if (probabilities[i]>randChoose){
				index = i; 
				break;
			}
		}
		if(index>=remainingFlights.size()) index = remainingFlights.size()-1;
		return index;
	}
	

	/**
	 * Return a list that contains all the remaining flights to choose that comply with all
	 * problem constrains. If the list is empty, then returns a random flight that departs
	 * from the main base. 
	 * 
	 * @param presentFlight: The present selected flight
	 * @param remainingFlights: List of remaining flights to select
	 * @param actualPairing: The pairing that is being constructed
	 * @return: A list containing all flights that comply with all the problem constrains
	 */
	private List<flight> getConstrainedFlights(flight presentFlight, List<flight> remainingFlights, List<flight> actualPairing ){

		boolean comply = true;
		int timeDifference=0;
		int flyingTime = 0;
		int workingTime = 0;
		List <flight> constrainedFlights = new ArrayList<flight>();
		
		for (int i = 0; i<actualPairing.size(); i++){
			
			if(i!=0){
				//Check whether crew has rest or not between flights
				timeDifference = actualPairing.get(i).returnDifferenceTime(actualPairing.get(i-1).getArrivalTime(),actualPairing.get(i-1).getArrivalDay());
				if (timeDifference>ACSP.MAX_SITTING_TIME_BETWEEN_FLIGHTS){
					workingTime -= timeDifference-ACSP.MAX_SITTING_TIME_BETWEEN_FLIGHTS;
				}
			} 

			flyingTime += actualPairing.get(i).getFlyingTime();
			workingTime += actualPairing.get(i).getFlyingTime();
		}
		
		for (flight x: remainingFlights){
			//Check all the constrains. Add only the flights that fit.
			
			//Basic constrains: Next flight must have departure on same or higher arrival day of the previous flight
			//					If the day is the same the departure time of the next flight must be higher
			if(presentFlight.getArrivalDay()>x.getDepartureDay()) comply = false;
			else if(presentFlight.getArrivalDay()==x.getDepartureDay() &&
					presentFlight.getArrivalTime()>x.getDepartureTime()) comply = false;
			
			//Constrain 1: Next flight must leave from the destination city of the previous flight
			else if (presentFlight.getDestinationCity().compareTo(x.getDepartureCity())!=0) comply = false;
			
			//Constrain 2: Max flights in a duty
			else if (actualPairing.size()+1>ACSP.MAX_FLIGHTS_IN_A_DUTY) comply = false;
			
			//Constrain 3: Maximum working time
			else if (workingTime+x.getFlyingTime()>ACSP.MAX_WORKING_TIME) comply = false;
			
			//Constrain 4: Total flying time in a duty 
			else if (x.getFlyingTime()+flyingTime>ACSP.MAX_FLYING_TIME) comply = false;
			
			//Constrain 5: Minimum sitting time between flights
			else if (x.returnDifferenceTime(presentFlight.getArrivalTime(), presentFlight.getArrivalDay())<ACSP.MIN_SITTING_TIME_BETWEEN_FLIGHTS) comply=false;
			
			//Constrain 6: Minimum rest time after working hours
			else if (workingTime>ACSP.MAX_WORKING_TIME && x.returnDifferenceTime(presentFlight.getArrivalTime(), presentFlight.getArrivalDay())<ACSP.MIN_REST_TIME_AFTER_MORE_THAN_12HOURS){
				comply = false;
			}
			
			if (comply){
				constrainedFlights.add(x);
			}
			comply = true;
		}
		return constrainedFlights;
	}
	
	

	/**
	 * Restart the ant solution containers
	 * @param newTrail:
	 * 		A matrix containing all the new values for pheromone trails 
	 */
	public void reboot(double [][] newTrail){
		for (int i=0; i<this.pairings.size(); i++){
			this.pairings.get(i).clear();
			}
			this.pairings.clear();
		for (int i=0;i<this.nodesFlights.length;i++){
			this.nodesFlights[i]=0;
		}
		this.trail = newTrail;
		this.finished=false;
	}
	
/******************************************HELPER FUNCTIONS***********************************************/
	/**
	 * Function that checks whether all flights have been selected for the problem
	 * @return:
	 * 		True if all flights have been selected. False elsewhere.
	 */
	public Boolean checkVisitedNodes(){
		for (int i=0; i<this.nodesFlights.length;i++){
			if(this.nodesFlights[i]!=1) return false;
		}
		return true;
	}
	
	/**
	 * Function to return the ID of the earliest flight to depart
	 * @param flights: 
	 * 		All the flights from where the earliest flight will be selected.
	 * @return:
	 * 		The ID of the earliest flight to depart
	 */
	public int getEarliestFlight(List<flight> flights){
		int earliestDay=100;
		int earliestTime=5000;
		int index = 0;
		for(int i=0; i<flights.size();i++){
			if(earliestDay>flights.get(i).getDepartureDay()){
				earliestDay = flights.get(i).getDepartureDay();
			} else if (earliestDay == flights.get(i).getDepartureDay() && earliestTime>flights.get(i).getDepartureTime()){
				earliestTime = flights.get(i).getDepartureTime();
				index = i;
			}
		}
		return index;
	}
	
	/**
	 * Function to return flights that have not been selected yet for the problem
	 * @return:
	 * 		List of pending to select flights
	 */
	public List<flight> getPendingFlights(){
		List<flight> penFlights = new ArrayList<flight>();
		int index=0;
		for (flight x: this.antFlights){
			if(this.nodesFlights[index]==0) penFlights.add(x);
			index++;
		}
		return penFlights;
	}
	
	/**
	 * Function to check whether an agent has finished its work
	 * @return:
	 * 		true when the agent has finished path construction
	 */
	public Boolean isFinished(){
		return this.finished;
	}
	
	
/*******************************************GETTERS*********************************************************/
	/**
	 * Getter to return the ID of the agent
	 * @return:
	 * 		The ID of the actual ant instance
	 */
	public int getAntID(){
		return this.antID;
	}
	
	/**
	 * Getter to return the pairings constructed by the agent
	 * @return:
	 * 		A List of List of flights that represent the pairings constructed
	 */
	public List<List<flight>> getPairings(){
		return this.pairings;
	}
	
	
}
