import java.util.ArrayList;
import java.util.List;

/**
 * 
 * @author Esteve Mir Esprrech
 * v1.0 November 2015
 * This class controls MMAS colony's behavior
 * 
 */
public class minMaxAntSystem {
	
	private double alpha; 				//Pheromone trail importance
	private double beta;  				//Heuristic function importance
	private double exTh;				//Exploration threshold. Exploration vs explotation
	private double evaporationP;		//Evaporation pheromone parameter
	private List<flight> flights;		//Flights in problem
	private double [][] trail;			//Trails of pheromones between edges
	private double [][] heuristic;		//Heuristic function between edges
	private double bestProblemCost;		//Best calculated cost
	private int ants;					//Number of ants
	private int maxIterations;			//Max iterations per ant
	private ant [] antColony;			//Array of ants
	private double minTrail;			//Minimum amount of pheromones on all arcs
	private double maxTrail;			//Maximum amount of pheromones on best arcs
	private double[][] infoAnts;		//Array to store results 
	private int actualIteration;		//Actual iteration
	
	/**
	 * Constructor
	 * 
	 * @param alpha:
	 * 		Pheromone modifier
	 * @param beta:
	 * 		Heuristic modifier
	 * @param exTh:
	 * 		Transition State function control parameter
	 * @param evaporation:
	 * 		Pheromone evaporation per iteration
	 * @param flights:
	 * 		List of flights for the problem
	 * @param ants:
	 * 		Number of agents
	 * @param it:
	 * 		Max number of iterations allowed
	 * @param initialTrail:
	 * 		Initial pheromone trail
	 */
	public minMaxAntSystem(double alpha, double beta, double exTh, double evaporation, List<flight> flights, int ants, int it, double initialTrail){
		this.alpha = alpha;
		this.beta = beta;
		this.exTh = exTh;
		this.evaporationP = evaporation;
		this.flights = flights;	
		this.ants = ants;
		this.maxIterations = it;
		this.infoAnts = new double[ants][it];
		this.actualIteration = 0;
		this.minTrail = 0;
		this.maxTrail = initialTrail;
		this.bestProblemCost = 99999999;

	}

	/***************************************************MAIN******************************************************/
	/**
	 * Init function for the algorithm
	 */
	public void init(){
		
		initTrailAndHeuristic();				//Initialize trail and heuristics matrix
		this.antColony = new ant[this.ants];	
		int iterationsBeforeStop = 200;			//Stop in 200 iterations if no more best results have been found
		int countWorseIterations = 0;
		double oldCost = 0;
		
		//Initialize ant colony
		for(int i=0; i<ants;i++){
			this.antColony[i] = new ant(i,this.alpha, this.beta, this.exTh, this.flights, this.trail, this.heuristic);
		}
		
		//While we don't reach maxIterations or we don't reach iterationsBeforeStop, do:
		for (int j=0; j<this.maxIterations; j++){
			
			//Init all the ants.
			for(int i=0; i<ants;i++){
				antColony[i].init();
				//while (!antColony[i].isFinished()); //Wait to avoid erroneous variable accesses.
			}
			
			oldCost = this.bestProblemCost;
			updatePheromoneTrail();					//Update pheromone trail with the results of the best ant.
			
			//Reboot all the ants.
			for(int i=0; i<this.ants;i++){
				this.antColony[i].reboot(trail);
			}
			this.actualIteration++;
			
			//Check iterationsBeforeStop
			if (this.bestProblemCost==oldCost){
				countWorseIterations+=1;
			} else {
				countWorseIterations = 0;
			}
			if(countWorseIterations>=iterationsBeforeStop) break;
		}
		//Print results
		calculateBestCost();
	}

	
/*************************************************FUNCTIONALITIES***********************************************/
	
	/**
	 * Initialize pheromone trails and heuristic parameters for each arc
	 */
	public void initTrailAndHeuristic(){
		
		int nFlights = this.flights.size();
		double tempDifference;
		flight tempFlight1;
		flight tempFlight2;
		
		this.trail = new double[nFlights][nFlights];
		this.heuristic = new double[nFlights][nFlights];
		
		//Initialize trail to maxTrail
		//Calculate heuristic between flights (1/Time_between_flights^(1/2)). Time is in hours
		for (int i=0;i<this.trail.length;i++){
			tempFlight1=this.flights.get(i);
			for (int j=0; j<this.trail[i].length;j++){
				this.trail[i][j]= this.maxTrail;
				tempFlight2=flights.get(j);
				tempDifference = tempFlight2.returnDifferenceTime(tempFlight1.getArrivalTime(), tempFlight1.getArrivalDay());
				
				if (tempFlight1.getDestinationCity().compareTo(tempFlight2.getDepartureCity())!=0){
					this.heuristic[i][j]=0;
				}
				else if (tempDifference<ACSP.MIN_DESC_TIME){
					this.heuristic[i][j]=0;
				}
				else {
					this.heuristic[i][j]=(1/Math.pow((tempDifference/60), 0.5)); 
				}
			}
		}
	}
	
	/**
	 * Function that updates pheromone trails. Acts in both evaporation and min-max processes.
	 */
	public void updatePheromoneTrail(){
		int actualAnt = calculateBestPath();
		double bestCost = this.infoAnts[actualAnt][this.actualIteration];
		ant tempAnt = this.antColony[actualAnt];
		List<List<flight>> pairings = new ArrayList<List<flight>>();
		
		//Update min and max values
		this.maxTrail = (1/(this.evaporationP*bestCost));
		this.minTrail = (this.maxTrail)/(tempAnt.getPairings().size()-1);

		//Update pheromone values on trails walked by the best ant
		pairings.clear();
		pairings = tempAnt.getPairings();
		for (int i=0; i<pairings.size();i++){
			for (int j=0; j<pairings.get(i).size();j++){
				if(j!=0){
					double increment = (100/(this.evaporationP*bestCost));
					this.trail[pairings.get(i).get(j-1).getID()][pairings.get(i).get(j).getID()] += increment;
				}
			}
		}
		
		//Calculate evaporation on pheromone trails
		for (int i=0;i<this.trail.length;i++){
			for (int j=0; j<this.trail[i].length;j++){
				this.trail[i][j]= this.trail[i][j]*(1-this.evaporationP);
				if (this.trail[i][j]<this.minTrail){
					this.trail[i][j] = this.minTrail;
				} else if (trail[i][j]>this.maxTrail){
					this.trail[i][j] = this.maxTrail;
				}
			}
		}
		
	}
	

/********************************************HELPER FUNCTIONS*****************************************************/
	
	/**
	 * Function that checks which ant has constructed the best path
	 * @return:
	 * 		The ID of the ant that has build the best path
	 */
	public int calculateBestPath(){
		List<List<flight>> pairings = new ArrayList<List<flight>>();
		double totalCost=0;
		double bestCost=9999999;
		int flyTime;
		int restTime;
		flight tempFlight1;
		flight tempFlight2;
		int actualAnt = 0;
		String city1, city2;
		
		for (ant a: this.antColony){
			pairings = a.getPairings();
			totalCost = 0;

			for (int i=0; i<pairings.size();i++){
				flyTime = 0;
				restTime = 0;
				totalCost+=pairings.get(i).size()*ACSP.BASIC_PAYMENT_PER_SERVICE;
				city1 = pairings.get(i).get(0).getDepartureCity();
				if (pairings.get(i).size()==1) totalCost+=5000; //Penalty for retrieved a individual solution
				for (int j=0; j<pairings.get(i).size();j++){
						tempFlight1 = pairings.get(i).get(j);
					if (j!=0){
						tempFlight2 = pairings.get(i).get(j-1);
						restTime = tempFlight1.returnDifferenceTime(tempFlight2.getArrivalTime(), tempFlight2.getArrivalDay());
						
						//Hotel must be paid if sitting time between flights is higher than MAX_SITTING_TIME_BETWEEN_FLIGHTS
						if (restTime>ACSP.MAX_SITTING_TIME_BETWEEN_FLIGHTS){
							totalCost+=ACSP.TOTAL_HOTEL_COST;
							if (restTime>ACSP.MIN_REST_TIME_AFTER_MORE_THAN_12HOURS){
								totalCost+=ACSP.BONUS_PER_EXTRA_MINUTE_OF_REST*(restTime-ACSP.MIN_REST_TIME_AFTER_MORE_THAN_12HOURS);
							}
						} else if (restTime>ACSP.MIN_SITTING_TIME_BETWEEN_FLIGHTS){
							totalCost+=ACSP.TOTAL_COST_PER_LOST_MINUTE;
						}
					}
					flyTime+= tempFlight1.getFlyingTime();
				}
				city2 = pairings.get(i).get(pairings.get(i).size()-1).getDestinationCity();
				if (city1.compareTo(city2)!=0) totalCost+=5000; //Penalty for retrieved a non complete solution
				if (flyTime>ACSP.MIN_PAID_FLY_TIME){
					totalCost+=ACSP.BONUS_PER_EXTRA_MINUTE_OF_FLIGHT*(flyTime-ACSP.MIN_PAID_FLY_TIME);
				}
			}
			if(totalCost<bestCost){
				bestCost = totalCost;
				if (bestCost<this.bestProblemCost){
					this.bestProblemCost=bestCost;
					
					
					 /*UNCOMMENT TO PRINT PAIRINGS
					pairings = a.getPairings();
					System.out.println("------------------------------------------------------------------------------------------");
						for (int i=0;i<pairings.size();i++){
							System.out.println("Pairing "+i);
							for(int j=0;j<pairings.get(i).size();j++){
								System.out.println("   "+pairings.get(i).get(j));
							}
						}
					System.out.println("------------------------------------------------------------------------------------------");
					*/
					
				}
				actualAnt = a.getAntID();
			}
			this.infoAnts[a.getAntID()][this.actualIteration] = bestCost;
		}
	  return actualAnt;
	}
	
	
	/**
	 * Function that uses all information saved during all the iterations to calculate
	 * statistics and print them to the console.
	 */
	public void calculateBestCost(){
		double cost = 0;
		double []iterationCost = new double[this.actualIteration];
		double bestIterationCost = 99999999;
		double worseIterationCost = 0;
		double bestIndividual = 99999999;
		double bestCost = 999999999;
		double worseCost = 0;
		for (int i=0; i<this.infoAnts.length;i++){
			cost=0;
			for (int j=0; j<this.infoAnts[i].length;j++){
				cost += this.infoAnts[i][j];
				if (this.infoAnts[i][j]<bestIndividual && this.infoAnts[i][j]!=0){
					bestIndividual = this.infoAnts[i][j];
				}
				
			}
			cost/=this.actualIteration;
			if(cost<bestCost)bestCost=cost;
			if(cost>worseCost) worseCost=cost;
		}
		for (int i=0; i<this.actualIteration;i++){
			for (int j=0; j<this.ants;j++){
				iterationCost[i] +=this.infoAnts[j][i];
			}
			iterationCost[i]/=this.ants;
			if(iterationCost[i]<bestIterationCost)bestIterationCost = iterationCost[i];
			if(iterationCost[i]>worseIterationCost)worseIterationCost = iterationCost[i];

		}

		System.out.print(this.bestProblemCost+",");
		System.out.print(bestIterationCost+",");
		System.out.print(worseIterationCost+",");
		System.out.print(bestCost+",");
		System.out.print(worseCost+",");
		System.out.print(this.actualIteration+",");
		
		
			
		//System.out.print("Best individual Cost is "+bestIndividual +" for Ant "+antWithBestIteration +" in iteration "+(iteration+1));
	}
	

}
